%{
/* Wolf - modified to read constraints. */

#include "ipp.h"
  
  /* default yywrap function - always treat EOF as an EOF  */
#define fctwrap() 1 
  
  char* rmdash( char* s )
    {
      s++; /* first letter is '-', skip it! */
      for( ; (*s == ' ') || (*s == '\t'); s++ )
	;
      return s;
    }

%}

a [Aa]
b [Bb]
c [Cc]
d [Dd]
e [Ee]
f [Ff]
g [Gg]
h [Hh]
i [Ii]
j [Jj]
k [Kk]
l [Ll]
m [Mm]
n [Nn]
o [Oo]
p [Pp]
q [Qq]
r [Rr]
s [Ss]
t [Tt]
u [Uu]
v [Vv]
w [Ww]
x [Xx]
y [Yy]
z [Zz]
DIGIT [[:digit:]]
POINT \.
NUMBER-NORM {DIGIT}+({POINT}{DIGIT}+)?
NUMBER-BARE {POINT}{DIGIT}+
NUMBER      ({NUMBER-BARE}|{NUMBER-NORM})
OP [+-*/]
COMP (">="|"<="|"=")
VAR \?[a-zA-Z][a-zA-Z0-9\-_]* 
NAME :?[a-zA-Z][a-zA-Z0-9\-_]* 
CONSTRAINT_VAR {VAR}\'?(\[({VAR}|{NAME})(,({VAR}|{NAME}))*\])?

%x COMMENT OVERREAD CONSTRAINT_SECTION

%%

"("[ \t]*{t}{e}{s}{t} {
  bracket_count = 1;
  BEGIN CONSTRAINT_SECTION;
  return(TEST_TOK);
}

"("[ \t]*{s}{e}{t}  {
  bracket_count = 1;
  BEGIN CONSTRAINT_SECTION;
  return(SET_TOK);
}

<CONSTRAINT_SECTION>{COMP} {
  strcpy(yylval.string, yytext);
  return(COMP_TOK);
}

"("  {  return('('); }

")"  {  return(')'); }

\([ \t]*{i}{n}"-"{p}{a}{c}{k}{a}{g}{e}  {  bracket_count = 1;
 BEGIN OVERREAD; }

":"{l}{e}{n}{g}{t}{h}  { return LENGTH_TOK; }

":"{s}{e}{r}{i}{a}{l}  { return SERIAL_TOK; }

":"{p}{a}{r}{a}{l}{l}{e}{l} { return PARALLEL_TOK; }

\([ \t]*":"{r}{e}{q}{u}{i}{r}{e}{m}{e}{n}{t}{s}  {  bracket_count = 1;
 BEGIN OVERREAD; }

":"{d}{o}{m}{a}{i}{n}"-"{v}{a}{r}{i}{a}{b}{l}{e}{s}  {  
  bracket_count = 1;
  BEGIN CONSTRAINT_SECTION;
  return(DOMAIN_VARIABLES_TOK); 
} 

{d}{e}{f}{i}{n}{e}  {  return(DEFINE_TOK); }

{p}{r}{o}{b}{l}{e}{m}  {  return(PROBLEM_TOK); }

{s}{i}{t}{u}{a}{t}{i}{o}{n}  {  return(SITUATION_TOK); }

":"{s}{i}{t}{u}{a}{t}{i}{o}{n}  {  return(BSITUATION_TOK); }

":"{o}{b}{j}{e}{c}{t}{s}  {  return(OBJECTS_TOK); }

":"{g}{o}{a}{l}  {  return(GOAL_TOK); }

":"{i}{n}{i}{t}  {  return(INIT_TOK); }

":"{d}{o}{m}{a}{i}{n}  {  return(BDOMAIN_TOK); }

\([ \t]*":"{e}{x}{t}{e}{n}{d}{s}  {  bracket_count = 1;
 BEGIN OVERREAD; }

{a}{n}{d}  {  return(AND_TOK); }

{i}{m}{p}{l}{y} {  return(IMPLY_TOK); }

{o}{r} {  return(OR_TOK); }

{f}{o}{r}{a}{l}{l} {  return(FORALL_TOK); }

{e}{x}{i}{s}{t}{s} {  return(EXISTS_TOK); }

{n}{o}{t}  {  return(NOT_TOK); }

<CONSTRAINT_SECTION>{NUMBER} {
  strcpy(yylval.string, yytext);
  return(NUMBER_TOK);
}

<CONSTRAINT_SECTION>{CONSTRAINT_VAR} {
  strcpy(yylval.string, strupcase(yytext));
  return(VARIABLE);
}

<CONSTRAINT_SECTION>"*" {
  strcpy(yylval.string, yytext);
  return('*');
}

<CONSTRAINT_SECTION>"/" {
  strcpy(yylval.string, yytext);
  return('/');
}

<CONSTRAINT_SECTION>"+" {
  strcpy(yylval.string, yytext);
  return('+');
}
<CONSTRAINT_SECTION>"-" {
  strcpy(yylval.string, yytext);
  return('-');
}

<CONSTRAINT_SECTION>"(" {
  bracket_count++;
  return('(');
}

<CONSTRAINT_SECTION>")" {
  bracket_count--;
  if (bracket_count == 0)
    BEGIN INITIAL;
  return(')');
}

"="  {  return(EQUAL_TOK); }

<CONSTRAINT_SECTION>"-"[ \t]*{f}{l}{u}{e}{n}{t} { 
  return(FLUENT_TOK); 
}

<CONSTRAINT_SECTION>"-"[ \t]*{f}{l}{o}{a}{t} { 
  return(FLOAT_TOK); 
}

:?[a-zA-Z][a-zA-Z0-9\-_]* {  
 strcpy(yylval.string, strupcase( yytext ) ); return(NAME); }

\?[a-zA-Z][a-zA-Z0-9\-_\[\]]* {  
 strcpy(yylval.string, strupcase( yytext ) ); return(VARIABLE); }

"-"[ \t]*[a-zA-Z][a-zA-Z0-9\-_\[\]]*  {   
 strcpy(yylval.string, rmdash(strupcase(yytext))); return(TYPE); }

"-"[ \t]*"("[ \t]*{e}{i}{t}{h}{e}{r} { return(EITHER_TOK); }

\;[^\n]*\n  {  lineno++; } 
\;[^\n]*  {  /* this will hold only in files that end with
		   a comment but no linefeed */ } 

<CONSTRAINT_SECTION>\;[^\n]*\n  {  lineno++; } 

<OVERREAD>\;[^\n]*\n  {  lineno++; } 

<COMMENT>(.^\")*\n    {  lineno++; }  ;

<INITIAL>\" { BEGIN COMMENT;}

<COMMENT>\" { BEGIN INITIAL;}

\n    {  lineno++; } 

<OVERREAD>(.^\(\))*\n  {  lineno++; }

<OVERREAD>[^\(\)]  {  }

<OVERREAD>\(  {  bracket_count++; }

<OVERREAD>\)  {  bracket_count--; 
  if (!bracket_count) BEGIN INITIAL; }

([1-9][0-9]*)|"0" {
  strcpy(yylval.string, yytext);
  return(NUMBER_TOK);
}

. {}

<CONSTRAINT_SECTION>. {}

<CONSTRAINT_SECTION>\n { lineno++; }

%%
