%{
/* Wolf - modified to handle constraints. */

#ifdef YYDEBUG
  extern int yydebug=1;
#endif


#include <stdio.h>
#include <string.h> 
#include "ipp.h"

#ifndef SCAN_ERR
#define SCAN_ERR
#define DEFINE_EXPECTED            0
#define PROBLEM_EXPECTED           1
#define PROBNAME_EXPECTED          2
#define LBRACKET_EXPECTED          3
#define RBRACKET_EXPECTED          4
#define DOMDEFS_EXPECTED           5
#define REQUIREM_EXPECTED          6
#define TYPEDLIST_EXPECTED         7
#define DOMEXT_EXPECTED            8
#define DOMEXTNAME_EXPECTED        9
#define TYPEDEF_EXPECTED          10
#define CONSTLIST_EXPECTED        11
#define PREDDEF_EXPECTED          12 
#define NAME_EXPECTED             13
#define VARIABLE_EXPECTED         14
#define ACTIONFUNCTOR_EXPECTED    15
#define ATOM_FORMULA_EXPECTED     16
#define EFFECT_DEF_EXPECTED       17
#define NEG_FORMULA_EXPECTED      18
#define NOT_SUPPORTED             19
#define SITUATION_EXPECTED        20
#define SITNAME_EXPECTED          21
#define BDOMAIN_EXPECTED          22
#define BADDOMAIN                 23
#define INIFACTS                  24
#define GOALDEF                   25
#define ADLGOAL                   26
  
#endif


static char *errmsg[] = {
  "'define' expected",
  "'problem' expected",
  "problem name expected",
  "'(' expected",
  "')' expected",
  "additional domain definitions expected",
  "requirements (e.g. ':strips') expected",
  "typed list of <%s> expected",
  "domain extension expected",
  "domain to be extented expected",
  "type definition expected",
  "list of constants expected",
  "predicate definition expected",
  "<name> expected",
  "<variable> expected",
  "action functor expected",
  "atomic formula expected",
  "effect definition expected",
  "negated atomic formula expected",
  "requirement %s not supported by this IPP version",  
  "'situation' expected",
  "situation name expected",
  "':domain' expected",
  "this problem needs another domain file",
  "initial facts definition expected",
  "goal definition expected",
  "first order logic expression expected",
  NULL
};


void fcterr( int errno, char *par );
int yyerror( char * msg );
int yylex();

%}

%start file

/* This may seem strange, but enables us to use the types known from
   ipp.h */
%union {
  char string[256];
  token tok;
  fact_list fl;
  op_list ol;
  token_list tl;
}


%type <tok> problem_name
%type <fl> adl_goal_description
%type <fl> adl_goal_description_star
%type <fl> literal_name_plus
%type <tl> literal_name
%type <tl> literal_term
%type <tl> atomic_formula_term
%type <tl> term_star
%type <tok> term
%type <tl> name_star
%type <tl> atomic_formula_name
%type <tok> predicate
%type <fl> typed_list_name
%type <fl> typed_list_variable
%type <tl> name_plus

%type <tl> constraint
%type <tl> cons_expr

%type <fl> typed_list_domain_var_decl
%type <tl> domain_var_decl
%type <tok> float_type


%token DEFINE_TOK
%token PROBLEM_TOK
%token SITUATION_TOK
%token BSITUATION_TOK
%token DOMAIN_VARIABLES_TOK
%token FLUENT_TOK /* These tokens are for */
%token FLOAT_TOK  /* constrained vars     */
%token <string> NUMBER_TOK
%token OBJECTS_TOK
%token BDOMAIN_TOK
%token INIT_TOK
%token GOAL_TOK
%token AND_TOK
%token NOT_TOK
%token <string> NAME
%token <string> VARIABLE
%token <string> TYPE
%token <string> COMP_TOK
%token TEST_TOK
%token SET_TOK
%token <string> NUMBER_TOK
%token LENGTH_TOK
%token PARALLEL_TOK
%token SERIAL_TOK
%token EQUAL_TOK
%token FORALL_TOK
%token IMPLY_TOK
%token OR_TOK
%token EXISTS_TOK
%token EITHER_TOK

%left '-' '+'
%left '/' '*'
%left NEG

%%

/* numbers in comments are only useful for the programmer
   and reader of the yacc source code. Each terminal or
   non-terminal symbol and (which is not nice) each
   program section in {brackets} has a number and may
   be reached by the $ operator - the numbers only
   help prevent counting all the time and producing
   ugly fat spots on the monitor */

/**********************************************************************/
file:
/* empty */
|
problem_definition file
;


/**********************************************************************/
problem_definition : 
'(' DEFINE_TOK         
{/* fcterr( PROBNAME_EXPECTED, NULL ); */}
problem_name
problem_defs
')'                 
{  
  strcpy(problemName, $4);
}
;


/**********************************************************************/
problem_name :
'(' PROBLEM_TOK
NAME       
')'        
{ 
  $$ = new_token( strlen($3)+1 );
  strcpy( $$, $3);
}
;


/**********************************************************************/
base_domain_name :
'(' BDOMAIN_TOK 
NAME       
')'
{ 
  if ( strcmp( $3, gdomain_name ) != SAME )
    {
	 /*      fcterr( BADDOMAIN, NULL );*/
      yyerror(NULL);
    }
}

/**********************************************************************/
problem_defs:
/* empty */
|
objects_def problem_defs
|
init_def problem_defs
|
goal_def problem_defs
|
base_domain_name problem_defs
|
length_def problem_defs
|
domain_variables_def problem_defs
;

/**********************************************************************
 * domain variable defn
/**********************************************************************/

domain_variables_def:
'(' DOMAIN_VARIABLES_TOK
typed_list_domain_var_decl
')'
{ 
  fact_list fl;

  if ($3) {
    for (fl = $3; fl->next; fl = fl->next) {
	 ;
    }
    fl->next = domain_vars_list;
    domain_vars_list = $3; 
  }
}
|
'(' DOMAIN_VARIABLES_TOK ')'
{ }
;

/* I'm not handling any type but fluent and float in this section! */
/* For now I'm not handling preset values */
typed_list_domain_var_decl:
domain_var_decl float_type
{
  token_list opt_num;

  $$ = new_fact_list();
  $$->item = $1;
  opt_num = $1->next;

  $$->item->next = new_token_list();
  $$->item->next->item = $2;

  $$->item->next->next = opt_num;

  $$->next = NULL;
}
|
domain_var_decl float_type typed_list_domain_var_decl   
/* end of list for one type */
{
  token_list opt_num;

  $$ = new_fact_list();
  $$->item = $1;
  opt_num = $1->next;

  $$->item->next = new_token_list();
  $$->item->next->item = $2;

  $$->item->next->next = opt_num;

  $$->next = $3;
}
|
domain_var_decl typed_list_domain_var_decl    
/* a list element (gets type from next one) */
{
  token_list opt_num;
  token s = $2->item->next->item;
  int len = strlen( s ) + 1;

  $$ = new_fact_list();
  $$->item = $1;
  opt_num = $1->next;

  $$->item->next = new_token_list();
  $$->item->next->item = new_token(len);
  strcpy($$->item->next->item, s);

  $$->item->next->next = opt_num;

  $$->next = $2;
}
;

domain_var_decl:
VARIABLE
{
  $$ = new_token_list();
  $$->item = new_token(strlen($1) + 1);
  strcpy($$->item, $1);
}
|
'(' VARIABLE NUMBER_TOK ')'
{
  $$ = new_token_list();
  $$->item = new_token(strlen($2) + 1);
  strcpy($$->item, $2);

  $$->next = new_token_list();
  $$->next->item = new_token(strlen($3) + 1);
  strcpy($$->next->item, $3);
}
;

float_type:
FLOAT_TOK
{ 
  $$ = new_token(strlen(FLOAT_STR) + 1); 
  strcpy($$, FLOAT_STR);
}
|
FLUENT_TOK
{ 
  $$ = new_token(strlen(FLUENT_STR) + 1); 
  strcpy($$, FLUENT_STR);
}
;

/**********************************************************************/
objects_def:
'(' OBJECTS_TOK
typed_list_name
')'
{ 
  fact_list f;
  type_tree root;
  type_tree_list ttl;

  if ( orig_constant_list )
    {
      for(f=orig_constant_list; f->next; f=f->next )
	;
      f->next = $3;
    }
  else
    orig_constant_list = $3;
  /* maybe new types are introduced here and must be added to the
     type tree */
  for ( f=orig_constant_list; f; f=f->next )
    {
      root = main_type_tree(); /* do not search the EITHER trees */
      if ( !find_branch( f->item->next->item, root ) )
	{ /* type of this constant is not known yet */
	  ttl = new_type_tree_list( f->item->next->item );
	  ttl->next = root->sub_types;
	  root->sub_types = ttl;
	}
    }
}
;

/**********************************************************************/
length_def:
'(' LENGTH_TOK
length_hints ')'
;

length_hints:
length_hint
length_hints
|
length_hint
;

length_hint:
'(' SERIAL_TOK NUMBER_TOK ')'
{
  gproblem_length_serial = atoi($3);
}
|
'(' PARALLEL_TOK NUMBER_TOK ')'
{
  gproblem_length_parallel = atoi($3);
}
;

/**********************************************************************/
init_def:
'(' INIT_TOK
{/* fcterr( INIFACTS, NULL ); */}
literal_name_plus
')'
{
  token_list temp;
  fact_list f;
  fact_list last = NULL;

  orig_initial_facts = $4;
  /* assuming a closed world we can ommit negated predicates in
     the initial state -- things not given do not hold anyway */
  for( f = orig_initial_facts; f; f = f->next )
    {
      if ( f->item->item[0] == '!' )
	{
	  if ( last == NULL )
	    {
	      orig_initial_facts = f->next;
	      /* f and its items should be freed now */
	    }
	  else
	    {
	      last->next = f->next;
	      /* f and its items should be freed now */
	    }
	}
      else
	{
	  /* Wolf - prepend an EFFECT_STR to each token */
	  temp = f->item;
	  f->item = new_token_list();
	  f->item->item = new_token(strlen(EFFECT_STR) + 1);
	  strcpy( f->item->item, EFFECT_STR );
	  f->item->next = temp;

	  /* update last */
	  last = f;
	}
    }
}
;

/**********************************************************************/
goal_def:
'(' GOAL_TOK
{ /*fcterr( GOALDEF, NULL );*/ }
adl_goal_description
')'
{
  /* Wolf - I'd rather have the goal as a fact_list */
  raw_goal_facts = $4;
}
;

/**********************************************************************
 * goal description providing full ADL
/**********************************************************************/

adl_goal_description: /* this provides only an easy-to-handle
			 structure for pl1-expressions which will
			 be preprocessed later. returns fact_list */
literal_term
{ 
  /* Wolf - the better to merge preconds and effects, I'm adding an initial
	precond/effect flag! */
  $$ = new_fact_list();
  $$->item = new_token_list();
  $$->item->item = new_token(strlen(PRECOND_STR) + 1);
  strcpy($$->item->item, PRECOND_STR);
  $$->item->next = $1;
}
|
'(' AND_TOK
adl_goal_description_star
')'
{ 
  fact_list f;
  
  $$ = make_adl_fact( AND_CONST );
  $$->next = $3; 
  for ( f=$$; f->next; f=f->next )
    ;
  f->next = make_adl_fact( RBRACK_CONST );
}
|
'(' OR_TOK
adl_goal_description_star
')'
{ 
  fact_list f;
  
  $$ = make_adl_fact( OR_CONST );
  $$->next = $3; 
  for ( f=$$; f->next; f=f->next )
    ;
  f->next = make_adl_fact( RBRACK_CONST );
}
|
'(' NOT_TOK
adl_goal_description
')'
{ 
  fact_list f;
  
  $$ = make_adl_fact( NOT_CONST );
  $$->next = $3; 
  for ( f=$$; f->next; f=f->next )
    ;
  f->next = make_adl_fact( ENDNOT_CONST );
}
|
'(' IMPLY_TOK
adl_goal_description
adl_goal_description
')'
{ 
  fact_list f;
  
  $$ = make_adl_fact( OR_CONST );
  $$->next = make_adl_fact( NOT_CONST );
  $$->next->next = $3; 
  for ( f=$$; f->next; f=f->next )
    ;
  f->next = make_adl_fact( ENDNOT_CONST);
  f->next->next = $4;
  for ( ; f->next; f=f->next )
    ;
  f->next = make_adl_fact( RBRACK_CONST );
}
|
'(' EXISTS_TOK
'('
typed_list_variable
')'
adl_goal_description
')'
{ 
  fact_list f;
  
  $$ = f = make_adl_fact( EXISTS_CONST );
  f->next = make_adl_fact( LBRACK_CONST );
  f->next->next = $4; 
  for ( ; f->next; f=f->next )
    ;
  f->next = make_adl_fact( RBRACK_CONST );
  f = f->next;
  f->next = $6;
  for ( ; f->next; f=f->next )
    ;
  f->next = make_adl_fact( RBRACK_CONST );
}
|
'(' FORALL_TOK
'('
typed_list_variable
')'
adl_goal_description
')'
{ 
  fact_list f;
  
  $$ = f = make_adl_fact( FORALL_CONST );
  f->next = make_adl_fact( LBRACK_CONST );
  f->next->next = $4; 
  for ( ; f->next; f=f->next )
    ;
  f->next = make_adl_fact( RBRACK_CONST );
  f = f->next;
  f->next = $6;
  for ( ; f->next; f=f->next )
    ;
  f->next = make_adl_fact( RBRACK_CONST );
}
;


adl_goal_description_star:
{
  $$ = NULL;
}
|
adl_goal_description adl_goal_description_star
{
  fact_list f;

  $$ = $1;
  if ( !$$ )
    $$ = $2;
  else
    {
      for ( f = $$; f->next; f = f->next )
	;
      f->next = $2;
    }
}

/**********************************************************************
 * some expressions used in many different rules
 **********************************************************************/

literal_term:
atomic_formula_term
{
  $$ = $1;
}
;

/**********************************************************************/

atomic_formula_term:
'('
predicate 
term_star
')'
{ 
  $$ = new_token_list();
  $$->item = new_token( strlen($2)+1 );
  strcpy( $$->item, $2 );
  $$->next = $3;
}
|
constraint
{
  $$ = $1;
}
;

/**********************************************************************/

term_star:
/* empty */
{ $$ = NULL; }
|
term
term_star
{
  $$ = new_token_list();
  $$->item = new_token( strlen($1)+1 );
  strcpy( $$->item, $1 );
  $$->next = $2;
}
;

/**********************************************************************/

term:
NAME
{ 
  $$ = new_token( strlen($1)+1 );
  strcpy( $$, $1 );
}
|
VARIABLE
{ 
  $$ = new_token( strlen($1)+1 );
  strcpy( $$, $1 );
}
;

/**********************************************************************/

name_plus:
NAME
{
  $$ = new_token_list();
  $$->item = new_token( strlen($1)+1 );
  strcpy( $$->item, $1 );
}
|
NAME name_plus
{
  $$ = new_token_list();
  $$->item = new_token( strlen($1)+1 );
  strcpy( $$->item, $1 );
  $$->next = $2;
}


/**********************************************************************/
typed_list_name:     /* returns fact_list */
/* empty */
{ $$ = NULL; }
|
NAME EITHER_TOK name_plus ')' typed_list_name
{ /* this is a very, very special case... it may mean that
     a. a parameter has one of the types in name_plus
     b. a type is subtype of one of the following types in name_plus 
     => for both possibilities we use the same solution:
     build a new type name: either_name1_name2_..._namen and check
     if a type with this name already exists. If so then NAME has this type,
     if not then build a new type, include it in the type tree and
     return NAME with the new type 
     The new artificial type is not a subtype of OBJECT because its own
     elements must already be instances of a subtype of OBJECT */
  token s;
  token_list t;
  type_tree tt;
  type_tree_list rootl, *st;
  
  s = new_token( MAX_LENGTH );
  strcpy( s, EITHER_STR );
  for ( t = $3; t; t = t->next )
    {
      strcat( s, CONNECTOR );
      strcat( s, t->item );
    }
  tt = NULL;
  for ( rootl = global_type_tree_list; rootl; rootl = rootl->next )
    if ( (tt = find_branch( s, rootl->item )) )
      break;
  if ( !tt )
    { /* the type doesn't exist yet, so build it */
      rootl = new_type_tree_list( s );
      rootl->next = global_type_tree_list;
      st = &(rootl->item->sub_types);
      for ( t = $3; t; t = t->next )
	{
	  if ( (tt = find_branch( t->item, main_type_tree() )) )
	    {
	      *st = new_type_tree_list( NULL );
	      (*st)->item = tt;
	      st = &((*st)->next);
	    }
	}  
    }

  /* now do the simple stuff: return a name with a (quite complicated)
     type */
  $$ = new_fact_list();
  $$->item = new_token_list();
  $$->item->item = new_token( strlen($1)+1 );
  strcpy( $$->item->item, $1 );
  $$->item->next = new_token_list();
  $$->item->next->item = s;
  $$->next = $5;
}
|
NAME TYPE typed_list_name   /* end of list for one type */
{
  $$ = new_fact_list();
  $$->item = new_token_list();
  $$->item->item = new_token( strlen($1)+1 );
  strcpy( $$->item->item, $1 );
  $$->item->next = new_token_list();
  $$->item->next->item = new_token( strlen($2)+1 );
  strcpy( $$->item->next->item, $2 );
  $$->next = $3;
}
|
NAME typed_list_name        /* a list element (gets type from next one) */
{
  $$ = new_fact_list();
  $$->item = new_token_list();
  $$->item->item = new_token( strlen($1)+1 );
  strcpy( $$->item->item, $1 );
  $$->item->next = new_token_list();
  if ( $2 )   /* another element (already typed) is following */
    {
      char* s = $2->item->next->item;
      int   l = strlen( s ) + 1;
      $$->item->next->item = new_token( l );
      strcpy( $$->item->next->item, s ); /* same type as the next one */
      $$->next = $2;
    }
  else /* no further element - it must be an untyped list */
    {
      $$->item->next->item = new_token( strlen(STANDARD_TYPE)+1 );
      strcpy( $$->item->next->item, STANDARD_TYPE );
      $$->next = $2;
    }
}
;

/***********************************************/

typed_list_variable:     /* returns fact_list */
/* empty */
{ $$ = NULL; }
|
VARIABLE EITHER_TOK name_plus ')' typed_list_variable
{ /* this is a very, very special case... it may mean that
     a parameter has one of the types in name_plus
     => build a new type name: either_name1_name2_..._namen and check
     if a type with this name already exists. If so then NAME has this type,
     if not then build a new type, include it in the type tree and
     return NAME with the new type 
     The new artificial type is not a subtype of OBJECT because its own
     elements must already be instances of a subtype of OBJECT */
  token s;
  token_list t;
  type_tree tt;
  type_tree_list rootl, *st;
  
  s = new_token( MAX_LENGTH );
  strcpy( s, EITHER_STR );
  for ( t = $3; t; t = t->next )
    {
      strcat( s, CONNECTOR );
      strcat( s, t->item );
    }
  tt = NULL;
  for ( rootl = global_type_tree_list; rootl; rootl = rootl->next )
    if ( (tt = find_branch( s, rootl->item )) )
      break;
  if ( !tt )
    { /* the type doesn't exist yet, so build it */
      rootl = new_type_tree_list( s );
      rootl->next = global_type_tree_list;
      st = &(rootl->item->sub_types);
      for ( t = $3; t; t = t->next )
	{
	  if ( (tt = find_branch( t->item, main_type_tree() )) )
	    {
	      *st = new_type_tree_list( NULL );
	      (*st)->item = tt;
	      st = &((*st)->next);
	    }
	}  
      global_type_tree_list = rootl;
    }

  /* now do the simple stuff: return a name with a (quite complicated)
     type */
  $$ = new_fact_list();
  $$->item = new_token_list();
  $$->item->item = new_token( strlen($1)+1 );
  strcpy( $$->item->item, $1 );
  $$->item->next = new_token_list();
  $$->item->next->item = s;
  $$->next = $5;
}
|
VARIABLE TYPE typed_list_variable   /* end of list for one type */
{
  $$ = new_fact_list();
  $$->item = new_token_list();
  $$->item->item = new_token( strlen($1)+1 );
  strcpy( $$->item->item, $1 );
  $$->item->next = new_token_list();
  $$->item->next->item = new_token( strlen($2)+1 );
  strcpy( $$->item->next->item, $2 );
  $$->next = $3;
}
|
VARIABLE typed_list_variable    /* a list element (gets type from next one) */
{
  $$ = new_fact_list();
  $$->item = new_token_list();
  $$->item->item = new_token( strlen($1)+1 );
  strcpy( $$->item->item, $1 );
  $$->item->next = new_token_list();
  if ( $2 )   /* another element (already typed) is following */
    {
      char* s = $2->item->next->item;
      int   l = strlen( s );
      $$->item->next->item = new_token( l+1 );
      strcpy( $$->item->next->item, s ); /* same type as the next one */
      $$->next = $2;
    }
  else /* no further element - it must be an untyped list */
    {
      $$->item->next->item = new_token( strlen(STANDARD_TYPE)+1 );
      strcpy( $$->item->next->item, STANDARD_TYPE );
      $$->next = $2;
    }
}
;


/**********************************************************************/

predicate:
NAME
{ 
  $$ = new_token( strlen($1)+1 );
  strcpy( $$, $1 );
}
|
EQUAL_TOK
{ 
  $$ = new_token( strlen(EQ_STR)+1 );
  strcpy( $$, EQ_STR );
}
;

/**********************************************************************/

literal_name_plus:
literal_name
{
  $$ = new_fact_list();
  $$->item = $1;
}
|
literal_name
literal_name_plus
{
   $$ = new_fact_list();
   $$->item = $1;
   $$->next = $2;
}
 
/**********************************************************************/

literal_name:
'(' NOT_TOK
atomic_formula_name
')'
{ 
  $$ = new_token_list();
  $$->item = new_token( strlen($3->item)+2 );
  strcpy( $$->item, "!" );
  strcat( $$->item, $3->item );
  $$->next = $3->next;
}
|
atomic_formula_name
{
  $$ = $1;
}
;

/**********************************************************************/

atomic_formula_name:
'('
predicate 
name_star
')'
{ 
  $$ = new_token_list();
  $$->item = new_token( strlen($2)+1 );
  strcpy( $$->item, $2 );
  $$->next = $3;
}
|
constraint
{
  $$ = $1;
}
;

/**********************************************************************/

name_star:
/* empty */
{ $$ = NULL; }
|
NAME
name_star
{
  $$ = new_token_list();
  $$->item = new_token( strlen($1)+1 );
  strcpy( $$->item, $1 );
  $$->next = $2;
}
;


/**********************************************************/
/*  CONSTRAINT STUFF                                      */
/**********************************************************/

constraint:
TEST_TOK
cons_expr COMP_TOK cons_expr ')'
{
  token_list tl;

  $$ = new_token_list();
  $$->item = new_token(strlen(TEST_STR) + 1);
  strcpy($$->item, TEST_STR);
  $$->next = $2;

  for (tl = $$->next; tl->next; tl = tl->next) {
    ;
  }

  tl->next = new_token_list();
  tl->next->item = new_token(strlen($3) + 1);
  strcpy(tl->next->item, $3);
  tl->next->next = $4;
}
|
SET_TOK
VARIABLE cons_expr ')'
{
  $$ = new_token_list();
  $$->item = new_token(strlen(SET_STR) + 1);
  strcpy($$->item, SET_STR);

  $$->next = new_token_list();
  $$->next->item = new_token(strlen($2) + 1);
  strcpy($$->next->item, $2);

  $$->next->next = new_token_list();
  $$->next->next->item = new_token(strlen("=") + 1);
  strcpy($$->next->next->item, "=");
  $$->next->next->next = $3;
}
;

cons_expr:
NUMBER_TOK
{
  $$ = new_token_list();
  $$->item = new_token(strlen($1) + 1);
  strcpy($$->item, $1);
}
|
VARIABLE
{
  $$ = new_token_list();
  $$->item = new_token(strlen($1) + 1);
  strcpy($$->item, $1);
}
|
cons_expr '*' cons_expr
{
  token_list tl;

  $$ = $1;
  for (tl = $$; tl->next; tl = tl->next) {
    ;
  }

  tl->next = new_token_list();
  tl->next->item = new_token(strlen("*") + 1);
  strcpy(tl->next->item, "*");
  tl->next->next = $3;
}
|
cons_expr '/' cons_expr
{
  token_list tl;

  $$ = $1;
  for (tl = $$; tl->next; tl = tl->next) {
    ;
  }

  tl->next = new_token_list();
  tl->next->item = new_token(strlen("/") + 1);
  strcpy(tl->next->item, "/");
  tl->next->next = $3;
}
|
cons_expr '+' cons_expr
{
  token_list tl;

  $$ = $1;
  for (tl = $$; tl->next; tl = tl->next) {
    ;
  }

  tl->next = new_token_list();
  tl->next->item = new_token(strlen("+") + 1);
  strcpy(tl->next->item, "+");
  tl->next->next = $3;
}
|
cons_expr '-' cons_expr
{
  token_list tl;

  $$ = $1;
  for (tl = $$; tl->next; tl = tl->next) {
    ;
  }

  tl->next = new_token_list();
  tl->next->item = new_token(strlen("-") + 1);
  strcpy(tl->next->item, "-");
  tl->next->next = $3;
}
|
'-' cons_expr %prec NEG
{
  $$ = new_token_list();
  $$->item = new_token(strlen("-") + 1);
  strcpy($$->item, "-");
  $$->next = $2;
}
|
'(' cons_expr ')'
{
  token_list tl = new_token_list();
  tl->item = new_token(strlen("(") + 1);
  strcpy(tl->item, "(");
  tl->next = $2;
  $$ = tl;

  for (tl = $$; tl->next; tl = tl->next) {
    ;
  }

  tl->next = new_token_list();
  tl->next->item = new_token(strlen(")") + 1);
  strcpy(tl->next->item, ")");
}
;


%%

#include "lex.fct.c"

/**********************************************************************
 * Functions
 **********************************************************************/

/* 
call	bison -pfct -bscan-fct scan-fct.y
*/
void fcterr( int errno, char *par )
{
  act_err = errno;
  if ( act_err_par )
    my_free( act_err_par );
  if ( par )
    {
      act_err_par = new_token( strlen(par)+1 );
      strcpy( act_err_par, par);
    }
  else
    act_err_par = NULL;
}

int yyerror( char *msg )
{
  fflush( stdout );
  fprintf(stderr,"\n%s: syntax error in line %d, '%s':\n", 
		act_filename, lineno, yytext );
  if ( act_err_par )
    {
      fprintf(stderr, errmsg[act_err], act_err_par );
      fprintf(stderr, "\n" );
    }
  else
    fprintf(stderr,"%s\n", errmsg[act_err] );

  OUTPUT_FILE;
  exit( 1 );
}


void load_fct_file( char *filename )
{
  FILE *fp;/* pointer to input files */

  /* open fact file */
  if( ( fp = fopen( filename, "r" ) ) == NULL ) {
    fprintf(stderr, "\nipp: can't find fact file: %s\n\n", filename );
    OUTPUT_FILE;
    exit( 1 );
  }
  act_filename = filename;
  lineno = 1; 
  yyin = fp;
  yyparse();
  fclose( fp );/* and close file again */
}

void print_factlist( fact_list list )
{

  fact_list i_list;
  token_list i_token;

  char * adl_strs[] = {"", "LIT", "AND", "OR", "IMPLY", "NOT", "EXISTS",
				   "FORALL", "[", "]", "VAR", "END_NOT"};

  for ( i_list = list; i_list; i_list = i_list->next ) 
    {
	 int c = get_adl_token(i_list);
	 if (c == VAR_CONST || c == LIT_CONST ||
		i_list->item->item[0] > ENDNOT_CONST) {
	   for ( i_token = i_list->item; i_token; i_token = i_token->next ) {
		printf( "%s ", i_token->item );
	   }
	 }
	 else {
	   printf("%s ", adl_strs[(int)i_list->item->item[0]]);
	 }
      printf( "\n" );
    }
}

void print_clauseelt( fact_list elt )
{
  token_list i_token;
  if (get_adl_token(elt) == LIT_CONST) {
    for ( i_token = elt->item; i_token; i_token = i_token->next )
	 printf( "%s ", i_token->item );    
  }
  else {
    switch (get_adl_token(elt)) {
    case AND_CONST:
	 printf("( . AND");
	 break;
    case OR_CONST:
	 printf("( . OR");
	 break;
    case RBRACK_CONST:
	 printf(")");
	 break;
    default:
	 printf("ERROR %d", get_adl_token(elt));
	 break;
    }
  }
}

void print_clauselist( fact_list list )
{

  fact_list i_list;

  printf("( . AND");

  for ( i_list = list; i_list; i_list = i_list->next ) 
    {
	 printf( " . " );
	 print_clauseelt(i_list);
    }
  printf(" . )");
}

void print_tokenlist(token_list list)
{
  token_list i_token;

  for ( i_token = list; i_token; i_token = i_token->next )
    {
      printf( "\n%s ", i_token->item );
      if (!strcmp(i_token->item, ""))
	printf("Match\n");
    }
}

